
"""
Core
"""


# Import from core_1
from crocodile.core_modules.core_1 import (
    validate_name,
    timestamp,
    str2timedelta,
    install_n_import,
    randstr,
    run_in_isolated_ve,
    Save
)

# Import from core_2
from crocodile.core_modules.core_2 import (
    Base,
    List,
    PLike
)

# Import from core_3
from crocodile.core_modules.core_3 import (
    Struct
)

# Import from core_4
from crocodile.core_modules.core_4 import (
    Display
)


def func(a: int, b: int) -> int:
    """
    A simple function that adds two integers.
    """
    return a + b


# Define public API
__all__ = [
    # From core_1
    'validate_name',
    'timestamp',
    'str2timedelta',
    'install_n_import',
    'randstr',
    'run_in_isolated_ve',
    'Save',

    # From core_2
    'Base',
    'List',
    'PLike',

    # From core_3
    'Struct',

    # From core_4
    'Display'
]

if __name__ == '__main__':
    pass
